use crate::{file_watcher::prelude::*, prelude::*};
use minijinja::Environment;
use pathdiff::diff_paths;
use std::{
    path::{Path, PathBuf},
    sync::Arc,
};
use thiserror::Error;
use tokio::sync::Mutex;
use tower_livereload::Reloader;

#[derive(Error, Debug)]
pub enum TemplateLoadError {
    #[error("template error: {0}")]
    Minijinja(#[from] minijinja::Error),

    #[error("io error: {0}")]
    Io(#[from] std::io::Error),
}

#[derive(Clone, Debug)]
pub struct Templates {
    env: Arc<Mutex<Environment<'static>>>,
    dir: PathBuf,
}

impl Templates {
    pub fn for_dir<P: Into<PathBuf>>(dir: P) -> Self {
        let env = Arc::new(Mutex::new(Environment::new()));
        Self {
            env,
            dir: dir.into(),
        }
    }

    pub async fn try_load<P: Into<PathBuf>>(dir: P) -> Result<Self, TemplateLoadError> {
        let result = Self::for_dir(dir);
        result.load_env().await?;
        Ok(result)
    }

    pub async fn start_watcher(
        self: Arc<Self>,
        reloader: Option<Reloader>,
    ) -> Result<Option<FileWatcher>, notify::Error> {
        if let Some(rl) = reloader {
            Ok(Some(self.watch(rl).await?))
        } else {
            Ok(None)
        }
    }

    async fn watch(self: Arc<Self>, reloader: Reloader) -> Result<FileWatcher, notify::Error> {
        let watcher = file_watcher(self.dir.clone(), move |ev| {
            futures::executor::block_on(async {
                if ev.kind.is_create() || ev.kind.is_modify() {
                    for p in ev.paths {
                        let _ = self.load_template(&p).await;
                    }
                    reloader.reload();
                }
            });
        })?;
        Ok(watcher)
    }

    #[instrument(skip(self, p))]
    pub async fn load_template<P>(&self, p: P) -> Result<(), TemplateLoadError>
    where
        P: AsRef<Path> + Copy,
    {
        let p = p.as_ref();
        if p.is_dir() {
            return Ok(());
        }
        let filename: String = diff_paths(p, "src/templates")
            .unwrap()
            .to_string_lossy()
            .into_owned();
        if [".bck", ".tmp"].iter().any(|s| filename.ends_with(s)) {
            debug!("skipping temporary file");
            return Ok(());
        }
        info!(filename);
        Ok(self
            .env
            .lock()
            .await
            .add_template_owned(filename, std::fs::read_to_string(p)?)?)
    }

    #[instrument(skip(self))]
    pub async fn load_env(&self) -> Result<(), TemplateLoadError> {
        for d in walkdir::WalkDir::new(&self.dir) {
            match d {
                Ok(d) => self.load_template(d.path()).await?,
                Err(_) => todo!(),
            }
        }
        Ok(())
    }

    pub async fn render<S: serde::ser::Serialize>(
        &self,
        template_name: &str,
        context: S,
    ) -> Result<String, minijinja::Error> {
        Ok(self
            .env
            .lock()
            .await
            .get_template(template_name)?
            .render(context)?)
    }
}
